#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

from .streams import (
    Activities,
    AdAccount,
    AdCreatives,
    Ads,
    AdSets,
    AdsInsights,
    AdsInsightsActionCarouselCard,
    AdsInsightsActionConversionDevice,
    AdsInsightsActionProductID,
    AdsInsightsActionReaction,
    AdsInsightsActionType,
    AdsInsightsActionVideoSound,
    AdsInsightsActionVideoType,
    AdsInsightsAgeAndGender,
    AdsInsightsCountry,
    AdsInsightsDeliveryDevice,
    AdsInsightsDeliveryPlatform,
    AdsInsightsDeliveryPlatformAndDevicePlatform,
    AdsInsightsDemographicsAge,
    AdsInsightsDemographicsCountry,
    AdsInsightsDemographicsDMARegion,
    AdsInsightsDemographicsGender,
    AdsInsightsDma,
    AdsInsightsPlatformAndDevice,
    AdsInsightsRegion,
    Campaigns,
    CustomAudiences,
    CustomConversions,
    Images,
    Videos,
)

__all__ = [
    "AdAccount",
    "AdCreatives",
    "Ads",
    "AdSets",
    "AdsInsights",
    "AdsInsightsActionType",
    "AdsInsightsAgeAndGender",
    "AdsInsightsCountry",
    "AdsInsightsDma",
    "AdsInsightsPlatformAndDevice",
    "AdsInsightsRegion",
    "AdsInsightsActionCarouselCard",
    "AdsInsightsActionConversionDevice",
    "AdsInsightsActionProductID",
    "AdsInsightsActionReaction",
    "AdsInsightsActionVideoSound",
    "AdsInsightsActionVideoType",
    "AdsInsightsDeliveryDevice",
    "AdsInsightsDeliveryPlatform",
    "AdsInsightsDeliveryPlatformAndDevicePlatform",
    "AdsInsightsDemographicsAge",
    "AdsInsightsDemographicsCountry",
    "AdsInsightsDemographicsDMARegion",
    "AdsInsightsDemographicsGender",
    "Campaigns",
    "CustomConversions",
    "CustomAudiences",
    "Images",
    "Videos",
    "Activities",
]



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#


import logging
from typing import Any, Dict, List, Mapping

from facebook_business.adobjects.adsinsights import AdsInsights

from airbyte_cdk import emit_configuration_as_airbyte_control_message
from airbyte_cdk.entrypoint import AirbyteEntrypoint
from airbyte_cdk.sources import Source
from airbyte_cdk.sources.message import InMemoryMessageRepository, MessageRepository
from source_facebook_marketing.spec import ValidAdSetStatuses, ValidAdStatuses, ValidCampaignStatuses


logger = logging.getLogger("airbyte_logger")


class MigrateAccountIdToArray:
    """
    This class stands for migrating the config at runtime.
    This migration is backwards compatible with the previous version, as new property will be created.
    When falling back to the previous source version connector will use old property `account_id`.

    Starting from `1.3.0`, the `account_id` property is replaced with `account_ids` property, which is a list of strings.
    """

    message_repository: MessageRepository = InMemoryMessageRepository()
    migrate_from_key: str = "account_id"
    migrate_to_key: str = "account_ids"

    @classmethod
    def should_migrate(cls, config: Mapping[str, Any]) -> bool:
        """
        This method determines whether the config should be migrated to have the new structure for the `custom_reports`,
        based on the source spec.
        Returns:
            > True, if the transformation is necessary
            > False, otherwise.
            > Raises the Exception if the structure could not be migrated.
        """
        return cls.migrate_from_key in config and cls.migrate_to_key not in config

    @classmethod
    def transform(cls, config: Mapping[str, Any]) -> Mapping[str, Any]:
        # transform the config
        config[cls.migrate_to_key] = [config[cls.migrate_from_key]]
        # return transformed config
        return config

    @classmethod
    def modify_and_save(cls, config_path: str, source: Source, config: Mapping[str, Any]) -> Mapping[str, Any]:
        # modify the config
        migrated_config = cls.transform(config)
        # save the config
        source.write_config(migrated_config, config_path)
        # return modified config
        return migrated_config

    @classmethod
    def migrate(cls, args: List[str], source: Source) -> None:
        """
        This method checks the input args, should the config be migrated,
        transform if neccessary and emit the CONTROL message.
        """
        # get config path
        config_path = AirbyteEntrypoint(source).extract_config(args)
        # proceed only if `--config` arg is provided
        if config_path:
            # read the existing config
            config = source.read_config(config_path)
            # migration check
            if cls.should_migrate(config):
                emit_configuration_as_airbyte_control_message(cls.modify_and_save(config_path, source, config))


class MigrateIncludeDeletedToStatusFilters(MigrateAccountIdToArray):
    """
    This class stands for migrating the config at runtime.
    This migration is backwards compatible with the previous version, as new property will be created.
    When falling back to the previous source version connector will use old property `include_deleted`.

    Starting from `1.4.0`, the `include_deleted` property is replaced with `ad_statuses`,
    `ad_statuses` and `campaign_statuses` which represent status filters.
    """

    migrate_from_key: str = "include_deleted"
    migrate_to_key: str = "ad_statuses"
    stream_filter_to_statuses: Mapping[str, List[str]] = {
        "ad_statuses": [status.value for status in ValidAdStatuses],
        "adset_statuses": [status.value for status in ValidAdSetStatuses],
        "campaign_statuses": [status.value for status in ValidCampaignStatuses],
    }

    @classmethod
    def should_migrate(cls, config: Mapping[str, Any]) -> bool:
        """
        This method determines whether the config should be migrated to have the new property for filters.
        Returns:
            > True, if the transformation is necessary
            > False, otherwise.
            > Raises the Exception if the structure could not be migrated.
        """
        config_is_updated = config.get(cls.migrate_to_key)
        no_include_deleted = not config.get(cls.migrate_from_key)
        return False if config_is_updated or no_include_deleted else True

    @classmethod
    def transform(cls, config: Mapping[str, Any]) -> Mapping[str, Any]:
        # transform the config
        for stream_filter, statuses in cls.stream_filter_to_statuses.items():
            config[stream_filter] = statuses
        # return transformed config
        return config


class MigrateSecretsPathInConnector:
    """
    This class stands for migrating the config at runtime.
    This migration is intended for backwards compatibility with the previous version, so existing secrets configurations gets migrated to new path.

    Starting from `2.2.0`, the `client_id`, `client_secret` and `access_token` will be placed at `credentials` path.
    """

    @classmethod
    def _should_migrate(cls, config: Mapping[str, Any]) -> bool:
        """
        This method determines whether the config should be migrated to nest existing fields at credentials.
        It is assumed if credentials does not exist on configuration, `client_id`, `client_secret` and `access_token` exists on root path.
        Returns:
            > True, if the migration is necessary
            > False, otherwise.
        """
        return "access_token" in config or "client_id" in config or "client_secret" in config

    @classmethod
    def migrate(cls, args: List[str], source: Source) -> None:
        """
        This method checks the input args, should the config be migrated,
        transform if neccessary and emit the CONTROL message.
        """
        # get config path
        config_path = AirbyteEntrypoint(source).extract_config(args)
        # proceed only if `--config` arg is provided
        if config_path:
            # read the existing config
            config = source.read_config(config_path)
            # migration check
            if cls._should_migrate(config):
                emit_configuration_as_airbyte_control_message(cls._modify_and_save(config_path, source, config))

    @classmethod
    def _transform(cls, config: Mapping[str, Any]) -> Mapping[str, Any]:
        # transform the config
        if "credentials" not in config:
            config["credentials"] = {
                "auth_type": "Service",
            }
        if "access_token" in config:
            config["credentials"]["access_token"] = config.pop("access_token")
        if "client_id" in config:
            config["credentials"]["auth_type"] = "Client"
            config["credentials"]["client_id"] = config.pop("client_id")
        if "client_secret" in config:
            config["credentials"]["auth_type"] = "Client"
            config["credentials"]["client_secret"] = config.pop("client_secret")
        # return transformed config
        return config

    @classmethod
    def _modify_and_save(cls, config_path: str, source: Source, config: Mapping[str, Any]) -> Mapping[str, Any]:
        # modify the config
        migrated_config = cls._transform(config)
        # save the config
        source.write_config(migrated_config, config_path)
        # return modified config
        return migrated_config


class RemoveActionReportTimeMigration:
    """
    Runtime config migrator that **removes** the deprecated
    ``action_report_time`` property.

    The field was deprecated starting in version 3.5.0.
    """

    migrate_key: str = "action_report_time"

    @classmethod
    def should_migrate(cls, config: Mapping[str, Any]) -> bool:
        """Return ``True`` when the deprecated key is present."""
        return any(cls.migrate_key in report for report in config.get("custom_insights", []))

    @classmethod
    def transform(cls, config: Mapping[str, Any]) -> Mapping[str, Any]:
        """
        Return a copy of *config* without the deprecated key.
        The original mapping is left untouched.
        """
        config_copy: Dict[str, Any] = dict(config)
        for report in config_copy["custom_insights"]:
            report.pop(cls.migrate_key, None)
        return config_copy

    @classmethod
    def modify_and_save(cls, config_path: str, source: Source, config: Mapping[str, Any]) -> Mapping[str, Any]:
        # modify the config
        migrated_config = cls.transform(config)
        # save the config
        source.write_config(migrated_config, config_path)
        # return modified config
        return migrated_config

    @classmethod
    def migrate(cls, args: List[str], source: Source) -> None:
        # get config path
        config_path = AirbyteEntrypoint(source).extract_config(args)
        # proceed only if `--config` arg is provided
        if config_path:
            # read the existing config
            config = source.read_config(config_path)
            # migration check
            if cls.should_migrate(config):
                emit_configuration_as_airbyte_control_message(cls.modify_and_save(config_path, source, config))


class MigrateDefaultActionBreakdowns:
    """
    This is temporary migration that will be removed in one week.
    This migration adds current default action breakdowns to existing configs
    to avoid breaking changes for existing users.
    """

    migrate_key: str = "default_ads_insights_action_breakdowns"

    @classmethod
    def should_migrate(cls, config: Mapping[str, Any]) -> bool:
        """Return ``True`` when the deprecated key is not present in config."""
        return cls.migrate_key not in config

    @classmethod
    def transform(cls, config: Mapping[str, Any]) -> Mapping[str, Any]:
        config[cls.migrate_key] = [
            AdsInsights.ActionBreakdowns.action_type,
            AdsInsights.ActionBreakdowns.action_target_id,
            AdsInsights.ActionBreakdowns.action_destination,
        ]
        return config

    @classmethod
    def modify_and_save(cls, config_path: str, source: Source, config: Mapping[str, Any]) -> Mapping[str, Any]:
        # modify the config
        migrated_config = cls.transform(config)
        # save the config
        source.write_config(migrated_config, config_path)
        # return modified config
        return migrated_config

    @classmethod
    def migrate(cls, args: List[str], source: Source) -> None:
        # get config path
        config_path = AirbyteEntrypoint(source).extract_config(args)
        # proceed only if `--config` arg is provided
        if config_path:
            # read the existing config
            config = source.read_config(config_path)
            # migration check
            if cls.should_migrate(config):
                emit_configuration_as_airbyte_control_message(cls.modify_and_save(config_path, source, config))



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

import http.client
import logging
import re
import sys
from datetime import timedelta
from typing import Any

import backoff
from facebook_business.exceptions import FacebookRequestError

from airbyte_cdk.models import FailureType
from airbyte_cdk.utils import AirbyteTracedException


# The Facebook API error codes indicating rate-limiting are listed at
# https://developers.facebook.com/docs/graph-api/overview/rate-limiting/
FACEBOOK_RATE_LIMIT_ERROR_CODES = {
    4,
    17,
    32,
    613,
    80000,
    80001,
    80002,
    80003,
    80004,
    80005,
    80006,
    80008,
}
FACEBOOK_TEMPORARY_OAUTH_ERROR_CODE = 2
FACEBOOK_BATCH_ERROR_CODE = 960
FACEBOOK_UNKNOWN_ERROR_CODE = 99
FACEBOOK_CONNECTION_RESET_ERROR_CODE = 104
DEFAULT_SLEEP_INTERVAL = timedelta(minutes=1)

logger = logging.getLogger("airbyte")


class JobException(Exception):
    """Scheduled job failed"""


class AccountTypeException(Exception):
    """Wrong account type"""


def retry_pattern(backoff_type, exception, **wait_gen_kwargs):
    def log_retry_attempt(details):
        _, exc, _ = sys.exc_info()
        logger.info(str(exc))
        logger.info(f"Caught retryable error after {details['tries']} tries. Waiting {details['wait']} more seconds then retrying...")

    def reduce_request_record_limit(details):
        _, exc, _ = sys.exc_info()
        # the list of error patterns to track,
        # in order to reduce the request page size and retry
        error_patterns = [
            "Please reduce the amount of data you're asking for, then retry your request",
            "An unknown error occurred",
        ]
        if (
            details.get("kwargs", {}).get("params", {}).get("limit")
            and exc.http_status() == http.client.INTERNAL_SERVER_ERROR
            and exc.api_error_message() in error_patterns
        ):
            # reduce the existing request `limit` param by a half and retry
            details["kwargs"]["params"]["limit"] = int(int(details["kwargs"]["params"]["limit"]) / 2)
            # set the flag to the api class that the last api call failed
            details.get("args")[0].last_api_call_is_successful = True
            # set the flag to the api class that the `limit` param was reduced
            details.get("args")[0].request_record_limit_is_reduced = True

    def revert_request_record_limit(details):
        """
        This method is triggered `on_success` after successful retry,
        sets the internal class flags to provide the logic to restore the previously reduced
        `limit` param.
        """
        # reference issue: https://github.com/airbytehq/airbyte/issues/25383
        # set the flag to the api class that the last api call was successful
        details.get("args")[0].last_api_call_is_successful = False
        # set the flag to the api class that the `limit` param is restored
        details.get("args")[0].request_record_limit_is_reduced = False

    def give_up(details):
        if isinstance(details["exception"], FacebookRequestError):
            raise traced_exception(details["exception"])

    def is_transient_cannot_include_error(exc: FacebookRequestError) -> bool:
        """After migration to API v19.0, some customers randomly face a BAD_REQUEST error (OAuthException) with the pattern:"Cannot include ..."
        According to the last comment in https://developers.facebook.com/community/threads/286697364476462/, this might be a transient issue that can be solved with a retry."""
        pattern = r"Cannot include .* in summary param because they weren't there while creating the report run."
        return bool(exc.http_status() == http.client.BAD_REQUEST and re.search(pattern, exc.api_error_message()))

    def should_retry_api_error(exc):
        if isinstance(exc, FacebookRequestError):
            call_rate_limit_error = exc.api_error_code() in FACEBOOK_RATE_LIMIT_ERROR_CODES
            temporary_oauth_error = exc.api_error_code() == FACEBOOK_TEMPORARY_OAUTH_ERROR_CODE
            batch_timeout_error = exc.http_status() == http.client.BAD_REQUEST and exc.api_error_code() == FACEBOOK_BATCH_ERROR_CODE
            unknown_error = exc.api_error_subcode() == FACEBOOK_UNKNOWN_ERROR_CODE
            connection_reset_error = exc.api_error_code() == FACEBOOK_CONNECTION_RESET_ERROR_CODE
            server_error = exc.http_status() == http.client.INTERNAL_SERVER_ERROR
            service_unavailable_error = exc.http_status() == http.client.SERVICE_UNAVAILABLE
            return any(
                (
                    exc.api_transient_error(),
                    unknown_error,
                    call_rate_limit_error,
                    batch_timeout_error,
                    is_transient_cannot_include_error(exc),
                    connection_reset_error,
                    temporary_oauth_error,
                    server_error,
                    service_unavailable_error,
                )
            )
        return True

    return backoff.on_exception(
        backoff_type,
        exception,
        jitter=None,
        on_backoff=[log_retry_attempt, reduce_request_record_limit],
        on_success=[revert_request_record_limit],
        on_giveup=[give_up],
        giveup=lambda exc: not should_retry_api_error(exc),
        **wait_gen_kwargs,
    )


def deep_merge(a: Any, b: Any) -> Any:
    """Merge two values, with `b` taking precedence over `a`."""
    if isinstance(a, dict) and isinstance(b, dict):
        # set of all keys in both dictionaries
        keys = set(a.keys()) | set(b.keys())

        return {key: deep_merge(a.get(key), b.get(key)) for key in keys}
    elif isinstance(a, list) and isinstance(b, list):
        return [*a, *b]
    elif isinstance(a, set) and isinstance(b, set):
        return a | b
    else:
        return a if b is None else b


FACEBOOK_CONFIG_ERRORS_TO_CATCH = [  # list of tuples (code, error_subcode)
    (100, 2446289),
]


def traced_exception(fb_exception: FacebookRequestError):
    """Add user-friendly message for FacebookRequestError

    Please see ../unit_tests/test_errors.py for full error examples
    Please add new errors to the tests
    """
    msg = fb_exception.api_error_message() or fb_exception.get_message()

    if "Invalid OAuth access token" in msg:
        failure_type = FailureType.config_error
        friendly_msg = (
            "The access token for this connection is invalid or corrupted. "
            "Please re-authenticate your Facebook connection in Airbyte. "
            "If re-authentication does not resolve the issue, go to facebook.com > Settings > Business Integrations, "
            "remove the Airbyte app, and then re-authenticate again."
        )

    elif "Error validating access token" in msg:
        failure_type = FailureType.config_error
        friendly_msg = "Invalid access token. Re-authenticate if FB oauth is used or refresh access token with all required permissions"

    elif "(#100) Missing permissions" in msg:
        failure_type = FailureType.config_error
        friendly_msg = (
            "Credentials don't have enough permissions. Check if correct Ad Account Id is used (as in Ads Manager), "
            "re-authenticate if FB oauth is used or refresh access token with all required permissions"
        )

    elif "permission" in msg:
        failure_type = FailureType.config_error
        friendly_msg = (
            "Credentials don't have enough permissions. Re-authenticate if FB oauth is used or refresh access token "
            "with all required permissions."
        )

    elif "An unknown error occurred" in msg and "error_user_title" in fb_exception._error:
        msg = fb_exception._error["error_user_title"]
        if "profile is not linked to delegate page" in msg or "el perfil no est" in msg:
            failure_type = FailureType.config_error
            friendly_msg = (
                "Current profile is not linked to delegate page. Check if correct business (not personal) "
                "Ad Account Id is used (as in Ads Manager), re-authenticate if FB oauth is used or refresh "
                "access token with all required permissions."
            )
    elif "reduce the amount of data" in msg:
        failure_type = FailureType.config_error
        friendly_msg = (
            "Please reduce the number of fields requested. Go to the schema tab, select your source, "
            "and unselect the fields you do not need."
        )
    elif "The start date of the time range cannot be beyond 37 months from the current date" in msg:
        failure_type = FailureType.config_error
        friendly_msg = "Please set the start date of your sync to be within the last 3 years."
    elif (fb_exception.api_error_code(), fb_exception.api_error_subcode()) in FACEBOOK_CONFIG_ERRORS_TO_CATCH:
        failure_type = FailureType.config_error
        friendly_msg = msg
    elif fb_exception.api_error_code() in FACEBOOK_RATE_LIMIT_ERROR_CODES:
        return AirbyteTracedException(
            message="The maximum number of requests on the Facebook API has been reached. See https://developers.facebook.com/docs/graph-api/overview/rate-limiting/ for more information",
            internal_message=str(fb_exception),
            failure_type=FailureType.transient_error,
            exception=fb_exception,
        )

    elif fb_exception.http_status() == 503:
        return AirbyteTracedException(
            message="The Facebook API service is temporarily unavailable. This issue should resolve itself, and does not require further action.",
            internal_message=str(fb_exception),
            failure_type=FailureType.transient_error,
            exception=fb_exception,
        )

    else:
        failure_type = FailureType.system_error
        error_code = fb_exception.api_error_code() if fb_exception.api_error_code() else fb_exception.http_status()
        friendly_msg = f"Error code {error_code}: {msg}."

    return AirbyteTracedException(
        message=friendly_msg or msg,
        internal_message=msg,
        failure_type=failure_type,
        exception=fb_exception,
    )
